from unittest import TestCase

import numpy as np

from scattertext import ScatterChart
from scattertext.ScatterChart import NoWordMeetsTermFrequencyRequirementsError
from scattertext.test.test_termDocMatrixFactory import build_hamlet_jz_term_doc_mat


class TestScatterChart(TestCase):
	def test_to_json(self):
		tdm = build_hamlet_jz_term_doc_mat()
		with self.assertRaises(NoWordMeetsTermFrequencyRequirementsError):
			ScatterChart(term_doc_matrix=tdm).to_dict('hamlet')
		j = (ScatterChart(term_doc_matrix=tdm,
		                  minimum_term_frequency=0)
		     .to_dict('hamlet'))
		self.assertEqual(set(j.keys()), set(['info', 'data']))
		self.assertEqual(set(j['info'].keys()),
		                 set(['not_category_name', 'category_name',
		                      'category_terms', 'not_category_terms']))
		expected = {"x": 0.0, "y": 0.42, "term": "art", "cat25k": 758, "ncat25k": 0, 's': 0.5}
		datum = j['data'][0]
		for var in ['x','y','cat25k','ncat25k', 's']:
			np.testing.assert_almost_equal(expected[var], datum[var])
		self.assertEqual(expected.keys(), datum.keys())
		self.assertEqual(expected['term'], datum['term'])



import json


class VizDataAdapter:
	def __init__(self, words_dict):
		self._word_dict = words_dict

	@property
	def word_dict(self):
		return self._word_dict

	@word_dict.setter
	def word_dict(self, val):
		self._word_dict = val

	@word_dict.deleter
	def word_dict(self):
		del self._word_dict

	def to_javascript(self):
		starter = 'function getDataAndInfo() { return'
		word_dict_json = json.dumps(self.word_dict)
		ender = '; }'
		return starter + word_dict_json + ender


import numpy as np
from scipy.stats import rankdata

from scattertext.TermDocMatrixFilter import filter_bigrams_by_pmis
from scattertext.Scalers import percentile_min, percentile_ordinal


class NoWordMeetsTermFrequencyRequirementsError(Exception):
	pass


class ScatterChart:
	def __init__(self,
	             term_doc_matrix,
	             minimum_term_frequency=3,
	             jitter=0,
	             seed=0):
		'''
		:param term_doc_matrix: TermDocMatrix
		:param jitter: float
		:param seed: float
		:return dataframe, html
		'''
		self.term_doc_matrix = term_doc_matrix
		self.jitter = jitter
		self.minimum_term_frequency = minimum_term_frequency
		self.seed = seed
		np.random.seed(seed)

	def to_dict(self,
	            category,
	            category_name=None,
	            not_category_name=None,
	            scores=None,
	            transform=percentile_ordinal):
		'''
		Returns a dictionary that encodes the scatter chart
		information. The dictionary can be dumped as a json document, and
		used in scattertext.html

		:param category: Category to annotate
		:param category_name: Name of category which will appear on web site.
		:param not_category_name: Name of non-category axis which will appear on web site.
		:param scores: Scores to use.  Default to Scaled F-Score.
		:param transform: Defaults to percentile_ordinal
		:return: dictionary {info: {category_name: ..., not_category_name},
		                     data: {term:, x:frequency [0-1], y:frequency [0-1],
		                            s: score,
		                            cat25k: freq per 25k in category,
		                            ncat25k: freq per 25k in non-category}}

		'''
		all_categories, other_categories = self._get_category_names(category)
		df = self._build_dataframe_for_drawing(all_categories, category, scores)
		df['x'], df['y'] = self._get_coordinates_from_transform_and_jitter_frequencies \
			(category, df, other_categories, transform)
		df['not cat freq'] = df[[x for x in other_categories]].sum(axis=1)
		json_df = df[['x', 'y', 'term']]
		json_df['cat25k'] = ((df[category + ' freq'] * 1. / df[category + ' freq'].sum()) * 25000)
		json_df['ncat25k'] = ((df['not cat freq'] * 1. / df['not cat freq'].sum()) * 25000)
		json_df['cat25k'] = json_df['cat25k'].apply(np.round).astype(np.int)
		json_df['ncat25k'] = json_df['ncat25k'].apply(np.round).astype(np.int)
		json_df['s'] = percentile_min(df['color_scores'])
		category_terms = list(json_df.sort_values('s')['term'][:10])
		not_category_terms = list(json_df.sort_values('s')['term'][:10])
		if category_name is None:
			category_name = category
		if not_category_name is None:
			not_category_name = 'Not ' + category_name
		j = {'info': {'category_name': category_name.title(),
		              'not_category_name': not_category_name.title(),
		              'category_terms': category_terms,
		              'not_category_terms': not_category_terms}}
		j['data'] = json_df.sort_values(by=['x', 'y', 'term']).to_dict(orient='records')
		return j

	def draw(self,
	         category,
	         num_top_words_to_annotate=4,
	         words_to_annotate=[],
	         scores=None,
	         transform=percentile_ordinal):
		try:
			import matplotlib.pyplot as plt
		except:
			raise Exception("matplotlib and mpld3 need to be installed to use this function.")
		try:
			from mpld3 import plugins, fig_to_html
		except:
			raise Exception("mpld3 need to be installed to use this function.")
		all_categories, other_categories = self._get_category_names(category)
		df = self._build_dataframe_for_drawing(all_categories, category, scores)
		x_data, y_data = self._get_coordinates_from_transform_and_jitter_frequencies \
			(category, df, other_categories, transform)
		df_to_annotate = df[(df['not category score rank'] <= num_top_words_to_annotate)
		                    | (df['category score rank'] <= num_top_words_to_annotate)
		                    | df['term'].isin(words_to_annotate)]
		words = list(df['term'])

		font = {'family': 'sans-serif',
		        'color': 'black',
		        'weight': 'normal',
		        'size': 'large'}

		fig, ax = plt.subplots()
		plt.figure(figsize=(10, 10))
		plt.gcf().subplots_adjust(bottom=0.2)
		plt.gcf().subplots_adjust(right=0.2)

		points = ax.scatter(x_data,
		                    y_data,
		                    c=-df['color_scores'],
		                    cmap='seismic',
		                    s=10,
		                    edgecolors='none',
		                    alpha=0.9)
		tooltip = plugins.PointHTMLTooltip(points,
		                                   ['<span id=a>%s</span>' % w for w in words],
		                                   css='#a {background-color: white;}')
		plugins.connect(fig, tooltip)
		ax.set_ylim([-.2, 1.2])
		ax.set_xlim([-.2, 1.2])
		ax.xaxis.set_ticks([0., 0.5, 1.])
		ax.yaxis.set_ticks([0., 0.5, 1.])
		ax.set_ylabel(category.title() + ' Frequency Percentile', fontdict=font, labelpad=20)
		ax.set_xlabel('Not ' + category.title() + ' Frequency Percentile', fontdict=font, labelpad=20)

		for i, row in df_to_annotate.iterrows():
			# alignment_criteria = row['category score rank'] < row['not category score rank']
			alignment_criteria = i % 2 == 0
			horizontalalignment = 'right' if alignment_criteria else 'left'
			verticalalignment = 'bottom' if alignment_criteria else 'top'
			term = row['term']
			ax.annotate(term,
			            (x_data[i], y_data[i]),
			            size=15,
			            horizontalalignment=horizontalalignment,
			            verticalalignment=verticalalignment,
			            )
		# texts.append(
		# ax.text(row['dem freq scaled'], row['rep freq scaled'], row['word'])
		# )
		# adjust_text(texts, arrowprops=dict(arrowstyle="->", color='r', lw=0.5))
		plt.show()
		return df, fig_to_html(fig)

	def _get_coordinates_from_transform_and_jitter_frequencies(self, category, df, other_categories, transform):
		x_data_raw = transform(df[other_categories].sum(axis=1))
		y_data_raw = transform(df[category + ' freq'])
		x_data = self._add_jitter(x_data_raw)
		y_data = self._add_jitter(y_data_raw)
		return x_data, y_data

	def _add_jitter(self, vec):
		'''
		:param vec: array to jitter
		:return:
		'''
		if self.jitter == 0:
			return vec
		else:
			to_ret = vec + np.random.rand(1, len(vec))[0] * self.jitter
			return to_ret

	def _build_dataframe_for_drawing(self, all_categories, category, scores):
		df = self.term_doc_matrix.get_term_freq_df()
		df['category score'] = np.array(self.term_doc_matrix.get_rudder_scores(category))
		df['not category score'] = np.sqrt(2) - np.array(self.term_doc_matrix.get_rudder_scores(category))
		df['color_scores'] = scores \
			if scores is not None \
			else np.array(self.term_doc_matrix.get_scaled_f_scores(category))
		df = filter_bigrams_by_pmis(
			df[df[all_categories].sum(axis=1) > self.minimum_term_frequency],
			threshold_coef=3)
		if len(df) == 0: raise NoWordMeetsTermFrequencyRequirementsError()
		df['category score rank'] = rankdata(df['category score'], method='ordinal')
		df['not category score rank'] = rankdata(df['not category score'], method='ordinal')
		df = df.reset_index()
		return df

	def _get_category_names(self, category):
		other_categories = [val + ' freq' for _, val \
		                    in self.term_doc_matrix._category_idx_store.items() \
		                    if val != category]
		all_categories = other_categories + [category + ' freq']
		return all_categories, other_categories



import pandas as pd
from numpy import nonzero

from scattertext import TermDocMatrix


class Corpus(TermDocMatrix):
	def __init__(self, X, y, term_idx_store, category_idx_store,
	             raw_texts, unigram_frequency_path=None):
		'''
		Parameters
		----------
		X csr_matrix
		y np.array
		term_idx_store IndexStore
		category_idx_store IndexStore
		raw_texts np.array or pd.Series
		unigram_frequency_path str or None
		'''
		TermDocMatrix.__init__(self, X, y, term_idx_store,
		                       category_idx_store, unigram_frequency_path)
		self._raw_texts = raw_texts

	def get_texts(self):
		'''
		Returns
		-------
		pd.Series, all raw documents
		'''
		return self._raw_texts

	def search(self, ngram):
		'''
		Parameters
		----------
		ngram str or unicode, string to search for

		Returns
		-------
		pd.DataFrame, {'texts': <matching texts>, 'categories': <corresponding categories>}

		'''
		mask = self._document_index_mask(ngram)
		return pd.DataFrame({
			'text': self._raw_texts[mask],
			'category': [self._category_idx_store.getval(idx)
			             for idx in self._y[mask]]
		})

	def search_index(self, ngram):
		'''
		Parameters
		----------
		ngram str or unicode, string to search for

		Returns
		-------
		np.array, list of matching document indices
		'''
		return nonzero(self._document_index_mask(ngram))[0]

	def _document_index_mask(self, ngram):
		idx = self._term_idx_store.getidxstrict(ngram.lower())
		mask = (self._X[:, idx] > 0).todense().A1
		return mask



from __future__ import print_function

import json
import os
import pkgutil
import re

from scattertext import Scalers
from scattertext.CSRMatrixTools import CSRMatrixFactory
from scattertext.IndexStore import IndexStore
from scattertext.ScatterChart import ScatterChart
from scattertext.TermDocMatrix import TermDocMatrix, InvalidScalerException
from scattertext.TermDocMatrixFactory import TermDocMatrixFactory, FeatsFromDoc
from scattertext.TermDocMatrixFromPandas import TermDocMatrixFromPandas
from scattertext.TermDocMatrixFilter import TermDocMatrixFilter, filter_bigrams_by_pmis
from scattertext.FastButCrapNLP import fast_but_crap_nlp
import scattertext.viz

def convention_speech_iter():
	relative_path = os.path.join('data', 'political_data.json')
	try:
		cwd = os.path.dirname(os.path.abspath(__file__))
		path = os.path.join(cwd, relative_path)
		return json.load(path)
	except:
		return json.loads(pkgutil.get_data('scattertext', relative_path))


def iter_party_convention_speech(nlp=None):
	if nlp is None:
		from spacy.en import English
		nlp = English()
	only_speaker_text_re = re.compile(
		r'((^|\n)((ANNOUNCER|AUDIENCE MEMBERS?): .+)($|\n)|(\n|^)((([A-Z\.()\- ]+): ))|\(.+\) *)',
		re.M)

	assert only_speaker_text_re.sub('', 'AUDIENCE MEMBERS: (Chanting.) USA! USA! USA! USA!') == ''
	assert only_speaker_text_re.sub('', 'AUDIENCE MEMBER: (Chanting.) USA! USA! USA! USA!') == ''
	assert only_speaker_text_re.sub('', 'ANNOUNCER: (Chanting.) USA! USA! USA! USA!') == ''
	assert only_speaker_text_re.sub('', 'TOM SMITH: (Chanting.) USA! USA! USA! USA!') == 'USA! USA! USA! USA!'
	assert only_speaker_text_re.sub('', 'DONALD TRUMP: blah blah blah!') == 'blah blah blah!'
	assert only_speaker_text_re.sub('', 'HILLARY CLINTON: (something parenthetical) blah blah blah!') == 'blah blah blah!'
	assert only_speaker_text_re.sub \
		       ('',
		        'ANNOUNCER: (Chanting.) USA! USA! USA! USA!\nTOM SMITH: (Chanting.) ONLY INCLUDE THIS! ONLY KEEP THIS! \nAUDIENCE MEMBER: (Chanting.) USA! USA! USA! USA!').strip() \
	       == 'ONLY INCLUDE THIS! ONLY KEEP THIS!'

	for category_struct in convention_speech_iter():
		category_name = category_struct['name']
		print('category', category_struct['name'])
		print('# speeches', len(category_struct['speeches']))
		for speech_raw in category_struct['speeches']:
			clean_speech = only_speaker_text_re.sub('', speech_raw)
			parsed_speech = nlp(clean_speech)
			yield category_name, parsed_speech



from unittest import TestCase

from scattertext.IndexStore import IndexStore

class TestIndexStore(TestCase):
	def test_main(self):
		index_store = IndexStore()
		self.assertEqual(index_store.getidx('a'), 0)
		self.assertEqual(index_store.getidx('b'), 1)
		self.assertEqual(index_store.getidx('a'), 0)
		self.assertEqual(index_store.getval(0), 'a')
		self.assertEqual(index_store.getval(1), 'b')
		self.assertTrue('a' in index_store)
		self.assertFalse('c' in index_store)
		self.assertEqual(set(index_store.values()), set(['a','b']))
		self.assertFalse(0 in index_store)
		self.assertTrue(index_store.hasidx(0))
		self.assertFalse(index_store.hasidx(2))
		self.assertEqual(index_store.getnumvals(), 2)
		self.assertEqual(list(index_store.items()), [(0, 'a'), (1, 'b')])

	def test_getidxstrict(self):
		index_store = IndexStore()
		self.assertEqual(index_store.getidx('a'), 0)
		self.assertEqual(index_store.getidx('b'), 1)
		self.assertEqual(index_store.getidx('a'), 0)
		with self.assertRaises(KeyError):
			index_store.getidxstrict('c')

	def test_batch_delete(self):
		index_store = IndexStore()
		self.assertEqual(index_store.getidx('a'), 0)
		self.assertEqual(index_store.getidx('b'), 1)
		self.assertEqual(index_store.getidx('c'), 2)
		self.assertEqual(index_store.getidx('d'), 3)
		with self.assertRaises(KeyError):
			new_idx_store = index_store.batch_delete_vals(['e', 'c'])
		new_idx_store = index_store.batch_delete_vals(['b','c'])
		self.assertEqual(new_idx_store.getidx('a'), 0)
		self.assertEqual(new_idx_store.getidx('c'), 2)
		self.assertEqual(new_idx_store.getidx('e'), 3)
		self.assertEqual(index_store.getidx('d'), 3)
		self.assertEqual(index_store.getidx('c'), 2)
		self.assertEqual(index_store.getidx('b'), 1)
		self.assertEqual(index_store.getidx('a'), 0)
		with self.assertRaises(ValueError):
			new_idx_store = index_store.batch_delete_idx([5, 1])
		new_idx_store = index_store.batch_delete_idx([2, 1])
		self.assertEqual(new_idx_store.getidx('a'), 0)
		self.assertEqual(new_idx_store.getidx('c'), 2)
		self.assertEqual(new_idx_store.getidx('e'), 3)

	def test_batch_delete_extra(self):
		index_store = IndexStore()
		self.assertEqual(index_store.getidx('a'), 0)
		self.assertEqual(index_store.getidx('b'), 1)
		self.assertEqual(index_store.getidx('c'), 2)
		self.assertEqual(index_store.getidx('d'), 3)
		self.assertEqual(index_store.getidx('e'), 4)
		self.assertEqual(index_store.getidx('f'), 5)
		del_idxstore = index_store.batch_delete_vals(['b', 'e'])
		self.assertEqual(list(del_idxstore.items()), [(0, 'a'), (1, 'c'), (2, 'd'), (3, 'f')])

		del_idxstore2 = del_idxstore.batch_delete_vals([])
		self.assertEqual(list(del_idxstore.items()), list(del_idxstore2.items()))



from unittest import TestCase

import numpy as np
import pandas as pd

from scattertext import fast_but_crap_nlp
from scattertext.Corpus import Corpus
from scattertext.CorpusFromPandas import CorpusFromPandas


def get_docs_categories():
	documents = [u"What art thou that usurp'st this time of night,",
	             u'Together with that fair and warlike form',
	             u'In which the majesty of buried Denmark',
	             u'Did sometimes march? by heaven I charge thee, speak!',
	             u'Halt! Who goes there?',
	             u'[Intro]',
	             u'It is I sire Tone from Brooklyn.',
	             u'Well, speak up man what is it?',
	             u'News from the East sire! THE BEST OF BOTH WORLDS HAS RETURNED!',
	             u'Speak up, speak up, this is a repeat bigram.'
	             ]
	categories = ['hamlet'] * 4 + ['jay-z/r. kelly'] * 5 + ['???']
	return categories, documents


class TestCorpusFromPandas(TestCase):
	def test_term_doc(self):
		self.assertIsInstance(self.corpus, Corpus)
		self.assertEqual(set(self.corpus.get_categories()),
		                 set(['hamlet', 'jay-z/r. kelly', '???']))
		self.assertEqual(self.corpus.get_num_docs(), 10)
		term_doc_df = self.corpus.get_term_freq_df()
		self.assertEqual(term_doc_df.ix['of'].sum(), 3)

	def test_get_texts(self):
		self.assertTrue(all(self.df['text'] == self.corpus.get_texts()))

	def test_search(self):
		expected = pd.DataFrame({'text': ["What art thou that usurp'st this time of night,",
		                                  "Together with that fair and warlike form"],
		                         'category': ['hamlet', 'hamlet']})
		self.assertIsInstance(self.corpus, Corpus)
		returned = self.corpus.search('that')
		np.testing.assert_array_equal(expected, returned)

	def test_search_bigram(self):
		expected = pd.DataFrame({'text': [u'Well, speak up man what is it?',
		                                  u'Speak up, speak up, this is a repeat bigram.'],
		                         'category': ['jay-z/r. kelly', '???']})
		self.assertIsInstance(self.corpus, Corpus)
		returned = self.corpus.search('speak up')
		np.testing.assert_array_equal(expected, returned)

	def test_search_index(self):
		expected = np.array([7, 9])
		self.assertIsInstance(self.corpus, Corpus)
		returned = self.corpus.search_index('speak up')
		np.testing.assert_array_equal(expected, returned)

	@classmethod
	def setUp(cls):
		categories, documents = get_docs_categories()
		cls.df = pd.DataFrame({'category': categories,
		                       'text': documents})
		cls.corpus = CorpusFromPandas(cls.df,
		                              'category',
		                              'text',
		                              nlp=fast_but_crap_nlp).build()



from unittest import TestCase

from scattertext import TermDocMatrixFilter
from scattertext.test.test_TermDocMat import get_hamlet_term_doc_matrix


class TestPMIFiltering(TestCase):
	def test_main(self):
		term_doc_mat = get_hamlet_term_doc_matrix()
		pmi_filter = TermDocMatrixFilter(pmi_threshold_coef=4, min_freq=3)
		filtered_term_doc_mat = pmi_filter.filter(term_doc_mat)
		print('term_doc_mat_df_len', len(term_doc_mat.get_term_freq_df()))
		print('filtered_term_doc_mat_df_len', len(filtered_term_doc_mat.get_term_freq_df()))
		print(term_doc_mat.get_term_freq_df())
		print(filtered_term_doc_mat.get_term_freq_df())
		self.assertLessEqual(len(filtered_term_doc_mat.get_term_freq_df()), len(term_doc_mat.get_term_freq_df()))


import numpy as np


def filter_bigrams_by_pmis(word_freq_df, threshold_coef=2):
	if len(word_freq_df.index) == 0:
		return word_freq_df

	low_pmi_bigrams = get_low_pmi_bigrams(threshold_coef, word_freq_df)
	return word_freq_df.drop(low_pmi_bigrams.index)


def get_low_pmi_bigrams(threshold_coef, word_freq_df):
	is_bigram = np.array([' ' in word for word in word_freq_df.index])
	unigram_freq = word_freq_df[~is_bigram].sum(axis=1)
	bigram_freq = word_freq_df[is_bigram].sum(axis=1)
	bigram_prob = bigram_freq / bigram_freq.sum()
	unigram_prob = unigram_freq / unigram_freq.sum()

	def get_pmi(bigram):
		return np.log(
			bigram_prob[bigram] / np.product([unigram_prob[word] for word in bigram.split(' ')])
		) / np.log(2)

	low_pmi_bigrams = bigram_prob[bigram_prob.index.map(get_pmi) < threshold_coef * 2]
	return low_pmi_bigrams


class TermDocMatrixFilter:
	def __init__(self, pmi_threshold_coef=2, min_freq=3):
		self._threshold_coef = pmi_threshold_coef
		self._min_freq = min_freq

	def filter(self, term_doc_matrix):
		'''
		:param term_doc_matrix: TermDocMatrix
		:return: TermDocMatrix pmi-filterd term doc matrix
		'''
		df = term_doc_matrix.get_term_freq_df()
		if len(df) == 0:
			return term_doc_matrix
		low_pmi_bigrams = get_low_pmi_bigrams(self._threshold_coef, df).index
		infrequent_terms = df[df.sum(axis=1) < self._min_freq].index
		return term_doc_matrix.remove_terms(set(low_pmi_bigrams | infrequent_terms))



import numpy as np

from scattertext import TermDocMatrixFromPandas
from scattertext.Corpus import Corpus


class CorpusFromPandas(TermDocMatrixFromPandas):
	'''Creates a Corpus from a pandas data frame.  A Corpus is a Term Document Matrix
	 preserves the original texts.

		Parameters
		----------
		data_frame : pd.DataFrame
			The data frame that contains columns for the category of interest
			and the document text.
		text_col : str
			The name of the column which contains the document text.
		category_col : str
			The name of the column which contains the category of interest.
		clean_function : function, optional
		nlp : function, optional
		use_lemmas : boolean, optional
		verbose : boolean, optional
			If true, prints a message every time a document index % 100 is 0.

		See Also
		--------
		TermDocMatrixFromPandas
		'''
	def _apply_pipeline_and_get_build_instance(self,
	                                           X_factory,
	                                           category_idx_store,
	                                           df,
	                                           parse_pipeline,
	                                           term_idx_store,
	                                           y):
		df.apply(parse_pipeline.parse, axis=1)
		X = X_factory.get_csr_matrix()
		y = np.array(y)
		raw_texts = df[self._text_col]
		return Corpus(X, y, term_idx_store, category_idx_store, raw_texts)


import string
from collections import Counter

import numpy as np

from scattertext.CSRMatrixTools import CSRMatrixFactory
from scattertext.IndexStore import IndexStore
from scattertext.TermDocMatrix import TermDocMatrix


class CategoryTextIterNotSetError(Exception):
	pass


class TermDocMatrixFactory:
	def __init__(self,
	             category_text_iter=None,
	             clean_function=lambda x: x,
	             nlp=None,
	             use_lemmas=False
	             ):
		"""Class for easy construction of a term document matrix.
		   This class let's you define an iterator for each document (text_iter),
		   an iterator for each document's category name (category_iter),
		   and a document cleaning function that's applied to each document
		   before it's parsed.

		   Parameters
		   ----------
		   category_text_iter : iter<str: category, unicode: document)>
		       An iterator of pairs. The first element is a string category
		       name, the second the text of a document.  You can also set this
		       using the function set_category_text_iter.
		   clean_function : function (default lambda x: x)
		       A function that takes a unicode document and returns
		       a cleaned version of that document
		   post_nlp_clean_function : function (default lambda x: x)
		       A function that takes a spaCy Doc
		   nlp : spacy.en.English (default None)
		       The spaCy parser used to parse documents.  If it's None,
		       the class will go through the expensive operation of
		       creating one to parse the text
		   use_lemmas : bool (default False)
		       Do we use the lower-cased strings or lemmas from
		        the spaCy tokenization?
		   Attributes
		   ----------
		   _clean_function : function
		       function that takes a unicode document and returns
		       a cleaned version of that document
		   _text_iter : iter<unicode>
		       an iterator that iterates through the unicode text of each
		        document
		   _category_iter : iter<str>
		       an iterator the same size as text iter that gives a string or
		       unicode name of each document catgory
		   Examples
		   --------
		   >>> import scattertext as ST
		   >>> documents = [u"What art thou that usurp'st this time of night,",
		    u'Together with that fair and warlike form',
		    u'In which the majesty of buried Denmark',
		    u'Did sometimes march? by heaven I charge thee, speak!',
				u'Halt! Who goes there?',
				u'[Intro]',
				u'It is I sire Tone from Brooklyn.',
				u'Well, speak up man what is it?',
				u'News from the East sire! THE BEST OF BOTH WORLDS HAS RETURNED!'
		    ]
		   >>> categories = ['hamlet'] * 4 + ['jay-z/r. kelly'] * 5
		   >>> clean_function = lambda text: '' if text.startswith('[') else text
		   >>> term_doc_mat = ST.TermDocMatrixFactory(
		    category_text_iter = zip(categories, documents),
		    clean_function = clean_function
		   ).build()

		"""
		self._category_text_iter = category_text_iter
		self._clean_function = clean_function
		self._nlp = nlp
		self._use_lemmas = use_lemmas
		self._entity_types_to_censor = set()

	def set_category_text_iter(self, category_text_iter):
		"""Initializes the category_text_iter

	   Paramters
	   ----------
	   category_text_iter : iter<str: category, unicode: document)>
		       An iterator of pairs. The first element is a string category
		       name, the second the text of a document.

		 Returns
		 ----------
		 self: TermDocMatrixFactory
		"""

		self._category_text_iter = category_text_iter
		return self

	def set_nlp(self, nlp):
		"""Adds a spaCy-compatible nlp function

	   Paramters
	   ----------
	   nlp : spacy.en.English

		 Returns
		 ----------
		 self: TermDocMatrixFactory
		"""

		self._nlp = nlp
		return self

	def build(self):
		"""Generate a TermDocMatrix from data in parameters.

		 Returns
		 ----------
		 term_doc_matrix : TermDocMatrix
		    The object that this factory class builds.
		"""
		if self._category_text_iter is None:
			raise CategoryTextIterNotSetError()
		nlp = self.get_nlp()

		category_document_iter = (
			(category, self._clean_function(raw_text))
			for category, raw_text
			in self._category_text_iter
		)
		term_doc_matrix = self._build_from_category_spacy_doc_iter(
			(
				(category, nlp(text))
				for (category, text)
				in category_document_iter
				if text.strip() != ''
			)
		)
		return term_doc_matrix

	def get_nlp(self):
		nlp = self._nlp
		if nlp is None:
			import spacy.en
			nlp = spacy.en.English()
		return nlp

	def censor_entity_types(self, entity_types):
		'''
		Entity types to exclude from feature construction. Terms matching
		specificed entities, instead of labeled by their lower case orthographic
		form or lemma, will be labeled by their entity type.

		Parameters
		----------
		entity_types : set of entity types outputted by spaCy
		  'TIME', 'WORK_OF_ART', 'PERSON', 'MONEY', 'ORG', 'ORDINAL', 'DATE',
		  'CARDINAL', 'LAW', 'QUANTITY', 'GPE', 'PERCENT'

		Returns
		---------
		self
		'''
		assert type(entity_types) == set
		self._entity_types_to_censor = entity_types
		return self

	def _build_from_category_spacy_doc_iter(self, category_doc_iter):
		'''
		Parameters
		----------
		category_doc_iter : iterator of (string category name, spacy.tokens.doc.Doc) pairs

		Returns
		----------
		t : TermDocMatrix
		'''
		term_idx_store = IndexStore()
		category_idx_store = IndexStore()
		X, y = self._get_features_and_labels_from_documents_and_indexes(category_doc_iter,
		                                                                category_idx_store,
		                                                                term_idx_store)
		return TermDocMatrix(X,
		                     y,
		                     term_idx_store=term_idx_store,
		                     category_idx_store=category_idx_store)

	def _get_features_and_labels_from_documents_and_indexes(self,
	                                                        category_doc_iter,
	                                                        category_idx_store,
	                                                        term_idx_store):
		y = []
		X_factory = CSRMatrixFactory()
		for document_index, (category, parsed_text) in enumerate(category_doc_iter):
			self._register_doc_and_category(X_factory,
			                                category,
			                                category_idx_store,
			                                document_index,
			                                parsed_text,
			                                term_idx_store, y)
		X = X_factory.get_csr_matrix()
		y = np.array(y)
		return X, y

	def _register_doc_and_category(self, X_factory, category, category_idx_store,
	                               document_index, parsed_text,
	                               term_idx_store, y):
		y.append(category_idx_store.getidx(category))
		document_features = self._get_features_from_parsed_text(parsed_text, term_idx_store)
		self._register_document_features_with_X_factory\
			(X_factory, document_index, document_features)


	def _register_document_features_with_X_factory(self, X_factory, doci, term_freq):
		for word_idx, freq in term_freq.items():
			X_factory[doci, word_idx] = freq

	def _get_features_from_parsed_text(self, parsed_text, term_idx_store):
		term_freq = Counter()
		for sent in parsed_text.sents:
			unigrams = []
			for tok in sent:
				if tok.pos_ not in ('PUNCT', 'SPACE', 'X'):
					if tok.ent_type_ in self._entity_types_to_censor:
						unigrams.append(tok.ent_type_)
					else:
						if self._use_lemmas:
							if tok.lemma_.strip():
								unigrams.append(tok.lemma_.strip())
						else:
							if tok.lower_.strip():
								unigrams.append(tok.lower_.strip())
			bigrams = list(map(' '.join, zip(unigrams[:-1], unigrams[1:])))
			self._augment_term_freq_with_unigrams_and_bigrams(bigrams, term_freq, term_idx_store, unigrams)
		return term_freq

	def _augment_term_freq_with_unigrams_and_bigrams(self, bigrams, term_freq, term_idx_store, unigrams):
		for term in unigrams + bigrams:
			term_freq[term_idx_store.getidx(term)] += 1


class FeatsFromDoc(TermDocMatrixFactory):
	def __init__(self, term_idx_store, clean_function=lambda x: x, nlp=None, use_lemmas=False, entity_types=set()):
		"""Class for extracting features from a new document.

	   Parameters
	   ----------
	   term_idx_store : IndexStore (index -> term)
	   clean_function : function (default lambda x: x)
	       A function that takes a unicode document and returns
	       a cleaned version of that document
	   post_nlp_clean_function : function (default lambda x: x)
	       A function that takes a spaCy Doc
	   nlp : spacy.en.English (default None)
	       The spaCy parser used to parse documents.  If it's None,
	       the class will go through the expensive operation of
	       creating one to parse the text
	   use_lemmas : bool (default False)
	       Do we use the lower-cased strings or lemmas from
	        the spaCy tokenization?
	   """
		TermDocMatrixFactory.__init__(self, clean_function=clean_function, nlp=nlp, use_lemmas=use_lemmas)
		self._term_idx_store = term_idx_store
		self._entity_types_to_censor = entity_types

	def feats_from_doc(self, raw_text):
		parsed_text = self._nlp(self._clean_function(raw_text))
		X_factory = CSRMatrixFactory()
		X_factory.set_last_col_idx(self._term_idx_store.getnumvals() - 1)
		term_freq = self._get_features_from_parsed_text(parsed_text, self._term_idx_store)
		self._register_document_features_with_X_factory(X_factory, 0, term_freq)
		return X_factory.get_csr_matrix()

	def _augment_term_freq_with_unigrams_and_bigrams(self, bigrams, term_freq, term_idx_store, unigrams):
		for term in unigrams + bigrams:
			if term in term_idx_store:
				term_freq[term_idx_store.getidx(term)] += 1


def build_from_category_whitespace_delimited_text(category_text_iter):
	'''
	:param category_text_iter: iterator of (string category name, one line per sentence, whitespace-delimited text) pairs
	:return: TermDocMatrix
	'''
	y = []
	X_factory = CSRMatrixFactory()
	term_idx_store = IndexStore()
	category_idx_store = IndexStore()
	for doci, (category, text) in enumerate(category_text_iter):
		y.append(category_idx_store.getidx(category))
		term_freq = Counter()
		for sent in text.strip(string.punctuation).lower().split('\n'):
			unigrams = []
			for tok in sent.strip().split():
				unigrams.append(tok)
			bigrams = list(map(' '.join, zip(unigrams[:-1], unigrams[1:])))
			for term in unigrams + bigrams:
				term_freq[term_idx_store.getidx(term)] += 1
		for word_idx, freq in term_freq.items():
			X_factory[doci, word_idx] = freq

	return TermDocMatrix(X=X_factory.get_csr_matrix(),
	                     y=np.array(y),
	                     term_idx_store=term_idx_store,
	                     category_idx_store=category_idx_store)
